/**
 * DBaseAuslesen ist eine Klasse, die DBase-Dateien, die typischerweise die
 * Extension DBF haben, lesen kann.
 */

import java.io.RandomAccessFile;
import java.io.FileNotFoundException;
import java.io.IOException;

public class DBaseAuslesen {

    private long recordCount = 0;
    private int recordSize = 0;
    private int offset = 0;

    public long getRecordCount() {
        return recordCount;
    }

    public int getRecordSize() {
        return recordSize;
    }

    public int getColumnCount() {
        return offset/32-1;
    }

    private String filename;

    private static final int DBFHEADER = 16;
    private static int unsign(byte wert) {
       return (wert < 0 ? 256 + wert : wert);
    }
 
    private class FieldDef {
        String name;
        char   type;
        int    length;
        int    dec;
        int    offset;
        void print() {
            System.out.println("Name: "+name+" Typ: "+type+" Len: "+length+" Dez: "+dec);
        }
    }

    FieldDef[] fieldDef;

    RandomAccessFile datei = null;

    public void close() {
        try {
            datei.close();
            datei = null;
        } catch(FileNotFoundException e) {
            System.out.println("Datei nicht vorhanden.");
            System.exit(1);
        } catch(IOException e) {
            System.out.println("Dateifehler.");
            System.exit(2);
        }
    }

    private byte [] dbfRecord = null;
    private long bufferedRecord = -1;

    private boolean readDbfRecord(long recordNr) {
        if (dbfRecord == null) {
            dbfRecord = new byte[recordSize];
        }
        if (bufferedRecord==recordNr) {
             return true;
        }
        try {
            datei.seek(recordNr*recordSize+offset);
            datei.read(dbfRecord);
            bufferedRecord = recordNr;
            return true;
        } catch(FileNotFoundException e) {
            System.out.println("Datei nicht vorhanden.");
            System.exit(1);
        } catch(IOException e) {
            System.out.println("Dateifehler.");
            System.exit(2);
        }
        return false;
    }

    private int getFieldIndex(String name) {
        int i=0;
        for (i=0; i<fieldDef.length; i++) {
             if (name.equals(fieldDef[i].name)) {
                 break;
             }
        }
        if (i>=fieldDef.length) {
            return -1;
        }
        return i;
    }

    public String getCharField(String name, long recordNr) {
        int fieldIndex = getFieldIndex(name);
        if (fieldIndex<0) {
            System.err.println("Feldname nicht gefunden");
            return null;
        }
        if (readDbfRecord(recordNr) ) {
            // wird in einen separaten Byte-Buffer uebertragen, um dann
            // eine Zeichensatzkonversion machen zu koennen
            byte[] buffer = new byte[fieldDef[fieldIndex].length];
            for (int j=0; j<fieldDef[fieldIndex].length; j++) {
               buffer[j] = dbfRecord[j+fieldDef[fieldIndex].offset];
            }
            try {
                String s = new String(buffer, "IBM850");
                return s;
            } catch (java.io.UnsupportedEncodingException ex) {
                System.err.println("DBase.getCharField: encoding problem");
            }
        }
        return null;
    }

    public double getNumField(String name, long recordNr) {
        return Double.parseDouble(getCharField(name, recordNr));
    }

    private void readTableDef() {
        fieldDef = new FieldDef[getColumnCount()];
        byte [] fieldBuffer = new byte[32];
        int offset = 1; // da gibt es eine Loeschmarkierung

        try {
            RandomAccessFile datei = new RandomAccessFile(filename, "rw");
            for (int i=0; i<getColumnCount(); i++) {
                datei.seek((i+1)*32);
                datei.read(fieldBuffer);
                StringBuffer sb = new StringBuffer();
                for (int j=0; j<10; j++) {
                    char zeichen = (char)fieldBuffer[j];
                    if (zeichen == 0) break;
                    sb.append(zeichen);
                }
                fieldDef[i] = new FieldDef();
                fieldDef[i].name = sb.toString();
                fieldDef[i].type  = (char)fieldBuffer[11];
                fieldDef[i].length = unsign(fieldBuffer[16]);
                fieldDef[i].dec = unsign(fieldBuffer[17]);
                fieldDef[i].offset = offset;
                offset += fieldDef[i].length;
                fieldDef[i].print();
            }
        } catch(FileNotFoundException e) {
            System.out.println("Datei nicht vorhanden.");
            System.exit(1);
        } catch(IOException e) {
            System.out.println("Dateifehler.");
            System.exit(2);
        }
    }

    public DBaseAuslesen(String file) {
        byte [] dbfHeader = new byte[DBFHEADER];
        try {
            datei = new RandomAccessFile(file, "rw");
            datei.read(dbfHeader);
            filename = file;
        } catch(FileNotFoundException e) {
            System.err.println("Datei nicht vorhanden.");
            System.exit(1);
        } catch(IOException e) {
            System.err.println("Dateifehler.");
            System.exit(2);
        }
        recordCount = unsign(dbfHeader[4]) +
                      unsign(dbfHeader[5])*0x100L +
                      unsign(dbfHeader[6])*0x10000L +
                      unsign(dbfHeader[7])*0x1000000L;
        // System.out.println("Anzahl der Records: " + recordCount);
        offset = unsign(dbfHeader[8]) +
                 unsign(dbfHeader[9])*0x100;
        // System.out.println("Offset: " + offset);
        recordSize = unsign(dbfHeader[10]) +
                             unsign(dbfHeader[11])*0x100;
        // System.out.println("Record-Groesse: " + recordSize);
        readTableDef();
    }

    static public void main(String[] args) {
        DBaseAuslesen db = new DBaseAuslesen("test.dbf");
        for (long i=0; i<db.getRecordCount(); i++) {
            String str = db.getCharField("DEUTSCH", i);
            String str2 = db.getCharField("UEBERSETZT", i);
            double zahl = db.getNumField("GEWUSST", i);
            System.out.println(str + "|" + str2 +"|"+ zahl);
        }
        db.close();
    }
}


